#TODO for version 1.0.0: Remove this file
import logging

from .common_hooks import additive_output_noise, hidden_patch_hook_fn

logging.warning("rome_hooks.py is deprecated and will be removed in version 1.0.0. Please use common_hooks.py instead.")

