# define some global variables
HF_MODELS = [
    'gpt2', 
    'gpt2-medium', 
    'gpt2-large', 
    'gpt2-xl',
    'gpt-neo-125M',
    'gpt-neo-1.3B',
    'gpt-neo-2.7B',
    'gpt-j-6b',
]

SESSION_STATE_VARIABLES = [
    'model',
    'tokenizer',
    'config',
    'registered_models',
    'registered_model_names',
    'num_layers',
]
